/*
	File:		TestMoreDisks.c

	Contains:	A program to exercise the MoreDisks module.

	Written by:	Quinn

	Copyright:	Copyright  1999 by Apple Computer, Inc., all rights reserved.

				You may incorporate this Apple sample source code into your program(s) without
				restriction. This Apple sample source code has been provided "AS IS" and the
				responsibility for its operation is yours. You are not permitted to redistribute
				this Apple sample source code as "Apple sample source code" after having made
				changes. If you're going to re-distribute the source, we require that you make
				it clear in the source that the code was descended from Apple sample source
				code, but that you've made changes.

	Change History (most recent first):

         <2>      7/5/99    Quinn   Added code to call and print the results of MoreIsDriveCDROM in
                                    DoListDriveQueue.
         <1>     16/3/99    Quinn   First checked in.
*/

/////////////////////////////////////////////////////////////////
// MoreIsBetter Setup

#include "MoreSetup.h"

// Mac OS Interfaces

#include <DriverGestalt.h>
#include <FSM.h>
#include <Disks.h>

// MIB Prototypes

#include "MoreInterfaceLib.h"
#include "MoreDisks.h"

// Standard C Interfaces

#include <stdio.h>
#include <stdlib.h>

/////////////////////////////////////////////////////////////////////
#pragma mark ----- Partition Map Tests -----

static OSErr PrintPartitionInfo(SInt16 drive)
{
    OSErr err;
    partInfoRec partitionInfo;

    err = MoreGetPartitionInfo(drive, &partitionInfo);
    if (err == noErr) {
        printf("MoreGetPartitionInfo succeeded!\n");
        printf("  partitionInfo.SCSIID = $%08lx\n", partitionInfo.SCSIID);
        printf("  partitionInfo.physPartitionLoc = %ld\n", partitionInfo.physPartitionLoc);
        printf("  partitionInfo.partitionNumber = %ld\n", partitionInfo.partitionNumber);

		#if 0
		{
		    SInt16 vRefNum;
	        
	        err = MoreGetPartitionVolume(MoreGetDriveRefNum(drive), &partitionInfo, &vRefNum);
	        if (err == noErr) {
	            printf("MoreGetPartitionVolume succeeded!\n");
	            printf("  vRefNum = %d\n", vRefNum);
	        } else {
	            printf("GetPartitionStatus failed with error %d.\n", err);
	            err = noErr;
	        }
	    }
		#endif

    }
    return err;
}

static OSErr ReadBlock(SInt16 drive, UInt32 blockNumber, void *blockBUffer)
{
    OSErr err;
    IOParam pb;
    
    pb.ioVRefNum = drive;
    pb.ioRefNum = MoreGetDriveRefNum(drive);
    pb.ioBuffer = blockBUffer;
    pb.ioReqCount = 512;
    pb.ioPosMode = fsFromStart;
    pb.ioPosOffset = blockNumber * 512;
    err = PBReadSync( (ParmBlkPtr) &pb );
    return err;
}

static void PrintBlock(UInt8 *block)
{
    UInt32 row;
    UInt32 col;
    
    for (row = 0; row < 32; row++) {
        for (col = 0; col < 16; col++) {
            printf("%02x ", block[row * 32 + col]);
        }
        printf("\n");
    }
}

static void PrintDDM(Block0 *ddm)
{
    SInt16 index;
    
    printf("PrintDDM\n");
    printf("\n");
    printf("sbSig = '%2.2s'\n", &ddm->sbSig);
    printf("sbBlkSize = %d\n", ddm->sbBlkSize);
    printf("sbBlkCount = %ld\n", ddm->sbBlkCount);
    printf("sbDevType = %d\n", ddm->sbDevType);
    printf("sbDevId = %d\n", ddm->sbDevId);
    printf("sbData = %ld\n", ddm->sbData);
    printf("sbDrvrCount = %d\n", ddm->sbDrvrCount);
    printf("\n");
    for (index = 0; index < ddm->sbDrvrCount; index++) {
        printf("  drivers[%d].ddBlock = %ld\n", index, ((DDMap *)(&ddm->ddBlock))[index].ddBlock);
        printf("  drivers[%d].ddSize = %d\n",   index, ((DDMap *)(&ddm->ddBlock))[index].ddSize);
        printf("  drivers[%d].ddType = $%04x\n",   index, ((DDMap *)(&ddm->ddBlock))[index].ddType);
    }
    printf("\n");
}

static void PrintPartitionMapEntry(Partition *part)
{
    printf("PrintPartitionMapEntry\n");
    printf("\n");
    printf("pmSig = '%2.2s'\n", &part->pmSig);
    printf("pmMapBlkCnt = %ld\n", part->pmMapBlkCnt);
    printf("pmPyPartStart = %ld\n", part->pmPyPartStart);
    printf("pmPartBlkCnt = %ld\n", part->pmPartBlkCnt);
    printf("pmPartName = %s\n", part->pmPartName);
    printf("pmParType = %s\n", part->pmParType);
    printf("pmLgDataStart = %ld\n", part->pmLgDataStart);
    printf("pmDataCnt = %ld\n", part->pmDataCnt);
    printf("pmPartStatus = %ld\n", part->pmPartStatus);
    printf("pmLgBootStart = %ld\n", part->pmLgBootStart);
    printf("pmBootSize = %ld\n", part->pmBootSize);
    printf("pmBootAddr = %ld\n", part->pmBootAddr);
    printf("pmBootAddr2 = %ld\n", part->pmBootAddr2);
    printf("pmBootEntry = %ld\n", part->pmBootEntry);
    printf("pmBootEntry2 = %ld\n", part->pmBootEntry2);
    printf("pmBootCksum = %08x\n", part->pmBootCksum);
    printf("pmProcessor = %s\n", part->pmProcessor);
    printf("\n");
}

static Block0    gDDM;
static Partition gPartition;

static OSStatus PrintEntirePartitionMap(SInt16 drive)
{
    OSStatus err;
    SInt16 workDrive;
    SInt32 partitionCount;
    SInt32 index;

	(void) PrintPartitionInfo(drive);
	    
    if ( MoreDriveSupportFileExchange(drive) ) {
        
        // Once we've created a cloned drive queue element, there's no
        // way to destroy it.  So the program just leaks drive queue elements.
        // To avoid this being an ongoing leak, before we create a new drive
        // queue element, we look for one that isn't being used by a volume
        // and reuse that.  Of course, before we use it we have to reset
        // it to point to the first two blocks on the disk.
        
        workDrive = MoreFirstDriveWithoutVolume(MoreGetDriveRefNum(drive));
        if ( workDrive == 0 ) {
	        err = MoreCreateNewDriveQueueElement(drive, 0, 2, &workDrive);
	    } else {
	    	err = MoreSetDrivePartition(workDrive, 0, 2);
        }
        
        // Now read and print the first disk block, ie the DDM.
        
        if (err == noErr) {
            err = ReadBlock(workDrive, 0, &gDDM);
        }
        if (err == noErr) {
            PrintDDM(&gDDM);
        }
        
        // Now read the second disk block, ie the first block of the
        // partition map.  Use that to determine the number of blocks
        // in the partition map and reset the drive to encompass the
        // entire partition map.
        
        if (err == noErr) {
            err = ReadBlock(workDrive, 1, &gPartition);
        }
        if (err == noErr) {
            err = MoreSetDrivePartition(workDrive, 0, gPartition.pmMapBlkCnt + 1);
        }
        
        // Now iterate through all the partition map entries, reading 
        // and printing each.
        
        if (err == noErr) {
            partitionCount = gPartition.pmMapBlkCnt;
            for (index = 1; index <= partitionCount; index++) {
                err = ReadBlock(workDrive, index, &gPartition);
                if (err == noErr) {
                    PrintPartitionMapEntry(&gPartition);
                }
                if (err != noErr) {
                    break;
                }
            }
        }
    
    } else {
        printf("Drive %d does not support PC Exchange.\n", drive);
        err = -1;
    }
    return err;
}

/////////////////////////////////////////////////////////////////////
#pragma mark ----- High Level Commands -----

static void DoListDriveQueue(void)
{
    SInt16 index;
    DrvQElPtr thisDrv;
    UInt32 sizeInBlocks;
    MoreDisksCDROMResponse isCDResponse;
    char cdChar;

	printf("List of drives:\n");
	printf("DrvNum Size\n");
	index = 1;
	do {
		thisDrv = MoreGetIndDrive(index);
		if (thisDrv != nil) {
			if ( MoreGetDriveSize(thisDrv->dQDrive, &sizeInBlocks) != noErr ) {
				sizeInBlocks = 0;
			}
			MoreIsDriveCDROM(thisDrv->dQDrive, &isCDResponse);
			switch (isCDResponse) {
				case kMoreDriveUnableToDetermineCDROM: cdChar = '?'; break;
				case kMoreDriveIsCDROM: cdChar = 'C'; break;
				case kMoreDriveIsNotCDROM: cdChar = 'N'; break;
				default:
					MoreAssertQ(false);
					break;
			}
			
			printf("%6d %8ld (%c)\n", thisDrv->dQDrive, sizeInBlocks, cdChar);
			index += 1;
		}
	} while (thisDrv != nil);
}

static void DoPrintPartitionMap(void)
{
	OSErr err;
    DrvQElPtr thisDrv;
    char driveStr[256];
    SInt16 driveNum;
	
	printf("Enter a drive number:\n");
	gets(driveStr);
	driveNum = atoi(driveStr);

	thisDrv = nil;
	if (driveNum > 0) {
		if ( MoreUTFindDriveQ(driveNum, &thisDrv) != noErr) {
			thisDrv = nil;
		}
	}
	if (thisDrv != nil) {
	    err = PrintEntirePartitionMap(thisDrv->dQDrive);
	} else {
		printf("Huh?\n");
		err = noErr;
	}
	
	if (err == noErr) {
		printf("Success!\n");
	} else {
		printf("Failed with error %d\n", err);
	}
}

static void DoFindFreeDriveQueueElement(void)
{
    char driveStr[256];
    SInt16 startNum;

	printf("Enter a drive number to start from (return for default, 5):");
	gets(driveStr);
	startNum = atoi(driveStr);
	if (startNum == 0) {
		startNum = 5;
	}
	printf("MoreFindFreeDriveNumber(%d) => %d\n", startNum, MoreFindFreeDriveNumber(startNum));
}

static void PrintHelp(void)
{
	printf("l) List drive queue elements\n");
	printf("p) Print partition map on drive\n");
	printf("f) Find a free drive queue element\n");
	printf("?) Print help\n");
	printf("q) Quit\n");
	printf("\n");
}

void main(void)
{
    char  commandStr[256];
    Boolean quitNow;
    
    printf("Hello Cruel World!\n");
    printf("QStandard.c\n");
    
    MoreAssertQ(sizeof(gDDM) == 512);
    MoreAssertQ(sizeof(gPartition) == 512);
	
	PrintHelp();
	quitNow = false;
	do {
		printf("Enter a command:\n");
		gets(commandStr);
		switch (commandStr[0]) {
			case 'l':
				DoListDriveQueue();
				break;
			case 'p':
				DoPrintPartitionMap();
				break;
			case 'f':
				DoFindFreeDriveQueueElement();
				break;
			case '?':
				PrintHelp();
				break;
			case 'q':
				quitNow = true;
				break;
			default:
				printf("Huh?\n");
				break;
		}
	} while ( ! quitNow );
	    
    printf("Done.  Press command-Q to Quit.\n");
}