/*
	File:		MorePreferences.cp

	Contains:	implements collection-based preferences container;
				supports both memory and disk

	Written by:	Pete Gontier

	Copyright:	Copyright (c) 1998 Apple Computer, Inc., All Rights Reserved.

				You may incorporate this Apple sample source code into your program(s) without
				restriction. This Apple sample source code has been provided "AS IS" and the
				responsibility for its operation is yours. You are not permitted to redistribute
				this Apple sample source code as "Apple sample source code" after having made
				changes. If you're going to re-distribute the source, we require that you make
				it clear in the source that the code was descended from Apple sample source
				code, but that you've made changes.

	Change History (most recent first):

         <3>     1/22/99    PCG     dunno
         <2>    11/11/98    PCG     fix header
         <1>    11/10/98    PCG     first big re-org at behest of Quinn

	Old Change History (most recent first):

         <4>     11/9/98    PCG     fix comment 3
         <3>     11/9/98    PCG     add copyright blurb and OpenPrefsResourceFile
         <2>     7/24/98    PCG	    UPP's for collection un/flattening
         <1>     6/16/98    PCG     initial checkin
*/

#include "MorePreferences.h"
#include "MoreProcesses.h"
#include "MoreCRC.h"

#include <Resources.h>
#include <Script.h>
#include <Collections.h>
#include <Folders.h>
#include <Errors.h>
#include <PLStringFuncs.h>

enum
{
	kPrefsFileMagicCookie	= 'ferp'
};

typedef struct
{
	UInt32			crc;
	UInt32			cookies [2];
	UInt8			flatCollectionData [ ];
}
tPrefsFormat, *tPrefsFormatP, **tPrefsFormatH;

typedef struct
{
	short			fileRefNum;
	tPrefsFormatP	prefsHeader;
}
tCollectionFlattenState, *tCollectionFlattenStateP;

typedef struct
{
	tPrefsFormatH	prefsData;
	UInt32			offset;
}
tCollectionUnflattenState, *tCollectionUnflattenStateP;

static Str31	gFolderName;
static OSType	gFileCreator;

////////////////////////////////////////////////////////////////////////////////////////////////////

pascal OSErr SetPrefsFolderName (ConstStr255Param folderName)
{
	OSErr err = noErr;

	if (MoreAssert (*folderName <= 31))
	{
		PLstrcpy (gFolderName,folderName);
	}

	return err;
}

pascal OSErr SetPrefsFileCreator (OSType ost)
{
	OSErr err = noErr;

	gFileCreator = ost;

	return err;
}

pascal OSErr GetPrefsFileCreator (OSType *ost)
{
	OSErr err = noErr;

	*ost = gFileCreator;

	return err;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

pascal OSErr NewPrefsCollection (PrefsCollection *newPrefs)
{
	*newPrefs = PrefsCollection (NewCollection ( ));
	if (!*newPrefs) return nilHandleErr;
	return noErr;
}

pascal OSErr DisposePrefsCollection (PrefsCollection doomedPrefs)
{
	DisposeCollection (Collection (doomedPrefs));
	return noErr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////

static pascal OSErr NewFSSpec (short vRefNum, long parID, ConstStr255Param path, FSSpecPtr *fssPP)
{
	OSErr err = noErr;

	if (!MoreAssert (*path <= 31))
		err = paramErr;
	else if (!(*fssPP = FSSpecPtr (NewPtr (sizeof (**fssPP)))))
		err = MemError ( );
	else
	{
		err = FSMakeFSSpec (vRefNum,parID,path,*fssPP);

		if (err && err != fnfErr)
		{
			DisposePtr (Ptr (*fssPP));
			*fssPP = nil;
		}
	}

	return err;	
}

static pascal OSErr NewHParamBlock (HParmBlkPtr &hpbp)
{
	hpbp = HParmBlkPtr (NewPtrClear (sizeof (*hpbp)));
	if (!hpbp) return MemError ( );
	return noErr;
}

static pascal OSErr NewCInfoParamBlock (CInfoPBPtr &cipbp)
{
	cipbp = CInfoPBPtr (NewPtrClear (sizeof (*cipbp)));
	if (!cipbp) return MemError ( );
	return noErr;
}

static pascal OSErr ScanDirectory
	(short vRefNum, long dirID, Str31 firstNameFound, OSType fdCreator, UInt16 &count)
{
	OSErr err = noErr;

	HParmBlkPtr hpbp;

	if (!(err = NewHParamBlock (hpbp)))
	{
		hpbp->fileParam.ioFDirIndex		= 1;
		hpbp->fileParam.ioVRefNum		= vRefNum;
		hpbp->fileParam.ioDirID			= dirID;
		hpbp->fileParam.ioNamePtr		= firstNameFound;

		count = 0;

		while (!(err = PBHGetFInfoSync (hpbp)))
		{
			if (kPrefsFileType == hpbp->fileParam.ioFlFndrInfo.fdType)
			if (fdCreator == hpbp->fileParam.ioFlFndrInfo.fdCreator)
			{
				hpbp->fileParam.ioNamePtr = nil;
				++count;
			}

			hpbp->fileParam.ioFDirIndex		+= 1;
			hpbp->fileParam.ioDirID			= dirID;
		}

		if (err == fnfErr)
			err = noErr;

		DisposePtr (Ptr (hpbp));
		if (!err) err = MemError ( );
	}

	return err;
}

static pascal OSErr GetPrefsSpec
	(ConstStr255Param fileName, OSType processSignature, Boolean forWriting, FSSpecPtr &fssP, UInt16 &count)
{
	OSErr	err = noErr;
	short	vRefNum;
	long	dirID;

	if (!(err = FindFolder (kOnSystemDisk,kPreferencesFolderType,kDontCreateFolder,&vRefNum,&dirID)))
	{
		if (*gFolderName)
		{
			CInfoPBPtr cipbp;

			if (!(err = NewCInfoParamBlock (cipbp)))
			{
				cipbp->dirInfo.ioNamePtr	= gFolderName;
				cipbp->dirInfo.ioVRefNum	= vRefNum;
				cipbp->dirInfo.ioDrDirID	= dirID;

				err = PBGetCatInfoSync (cipbp);

				if (err == fnfErr)
				{
					//
					//	If we could not find the deeper folder, and we
					//	we are attempting to read the file, we should
					//	scan the main Prefs folder. If we're writing,
					//	then create the folder.
					//

					if (!forWriting)
						err = noErr;
					else
					{
						long newDirID;

						if (!(err = DirCreate (vRefNum, dirID, gFolderName, &newDirID)))
							dirID = newDirID;
					}
				}
				else if (ioDirMask & cipbp->dirInfo.ioFlAttrib)
				{
					//
					//	We found the deeper folder; scan it instead of the Preferences folder.
					//

					dirID = cipbp->dirInfo.ioDrDirID;
				}
				else
				{
					//
					//	If we found a file instead of a folder with the designated name,
					//	scan the main Prefs folder. (Do nothing here; err == noErr.)
					//
				}

				DisposePtr (Ptr (cipbp));
				(void) MoreAssert (noErr == MemError ( ));
			}
		}

		if (!err)
		{
			Str31 firstNameFound;

			if (!(err = ScanDirectory (vRefNum,dirID,firstNameFound,processSignature,count)))
			{
				if (!count)
					err = NewFSSpec (vRefNum,dirID,fileName,&fssP);
				else if (count == 1)
					err = NewFSSpec (vRefNum,dirID,firstNameFound,&fssP);
				else if (fileName)
					err = NewFSSpec (vRefNum,dirID,fileName,&fssP);
				else
					err = fnfErr;
			}
		}
	}

	return err;
}

pascal OSErr GetNewPrefsCollection (PrefsCollection *newPrefs, ConstStr255Param fileName)
{
	OSErr err = noErr;

	OSType prefsFileCreator;

	if (gFileCreator)
		prefsFileCreator = gFileCreator;
	else
	{

		ProcessInfoRec pir;

		if (!(err = GetSomeProcessInfo (nil,&pir)))
			prefsFileCreator = pir.processSignature;
	}

	if (!err)
	{
		FSSpecPtr	fssP;
		UInt16		count;

		err = GetPrefsSpec (fileName,prefsFileCreator,false,fssP,count);

		if (!err)
			err = GetNewPrefsCollectionFromFile (newPrefs,fssP);

		if (fssP)
		{
			DisposePtr (Ptr (fssP));
			if (!err) err = MemError ( );
		}
	}

	return err;
}

pascal OSErr WritePrefsCollection (PrefsCollection pc, ConstStr255Param fileName)
{
	OSErr err = noErr;

	OSType prefsFileCreator;

	if (gFileCreator)
		prefsFileCreator = gFileCreator;
	else
	{

		ProcessInfoRec pir;

		if (!(err = GetSomeProcessInfo (nil,&pir)))
			prefsFileCreator = pir.processSignature;
	}

	if (!err)
	{
		FSSpecPtr	fssP;
		UInt16		count;
		Boolean		createdItHere = false;

		err = GetPrefsSpec (fileName,prefsFileCreator,true,fssP,count);

		if (err == fnfErr && !count)
		{
			err = FSpCreate (fssP,prefsFileCreator,kPrefsFileType,smSystemScript);
			if (!err) createdItHere = true;
		}

		if (!err)
		{
			err = WritePrefsCollectionToFile (pc,fssP);

			if (err && createdItHere)
				(void) FSpDelete (fssP);
		}

		if (fssP)
		{
			DisposePtr (Ptr (fssP));
			if (!err) err = MemError ( );
		}
	}

	return err;
}

static pascal OSErr CollectionUnflattenProc (SInt32 size, void *data, void *refCon)
{
	tCollectionUnflattenStateP state = tCollectionUnflattenStateP (refCon);
	BlockMove (state->offset + (**(state->prefsData)).flatCollectionData, data, size);
	state->offset += size;
	return noErr;
}

pascal OSErr GetNewPrefsCollectionFromFile (PrefsCollection *pc, const FSSpec *fssP)
{
	OSErr err = noErr;

	short fileRefNum;

	if (!(err = FSpOpenDF (fssP,fsRdPerm,&fileRefNum)))
	{
		Size eof;

		if (!(err = GetEOF (fileRefNum, &eof)))
		{
			Handle crcBuffer = NewHandle (eof);

			if (!crcBuffer)
				err = MemError ( );
			else
			{
				HLockHi (crcBuffer);
				if (!(err = MemError ( )))
				{
					if (!(err = FSRead (fileRefNum,&eof,*crcBuffer)))
					{
						tPrefsFormatH flatPrefs = tPrefsFormatH (crcBuffer);

						if ((**flatPrefs).cookies [0] != kPrefsFileMagicCookie)
							err = badFileFormat;
						else if ((**flatPrefs).cookies [1] != kPrefsFileMagicCookie)
							err = badFileFormat;
						else
						{
							UInt32 crc = MoreCRC32 (0, (**flatPrefs).flatCollectionData, eof - sizeof (tPrefsFormat));

							if ((**flatPrefs).crc != crc)
								err = badFileFormat;
							else if (!(err = NewPrefsCollection (pc)))
							{
								HUnlock (crcBuffer);
								if (!(err = MemError ( )))
								{
									CollectionFlattenUPP upp = NewCollectionFlattenProc (CollectionUnflattenProc);

									if (!upp)
										err = MemError ( );
									else
									{
										tCollectionUnflattenState unflattenState = { flatPrefs, 0 };
										err = UnflattenCollection (Collection (*pc),upp,&unflattenState);
										DisposeRoutineDescriptor (upp);
									}
								}

								if (err) DisposePrefsCollection (*pc);
							}
						}						
					}
				}

				DisposeHandle (crcBuffer);
				if (!err) err = MemError ( );
			}
		}

		OSErr err2 = FSClose (fileRefNum);
		if (!err) err = err2;
	}

	return err;
}

static pascal OSErr CollectionFlattenProc (SInt32 size, void *data, void *refCon)
{
	tCollectionFlattenStateP state = tCollectionFlattenStateP (refCon);
	state->prefsHeader->crc = MoreCRC32 (state->prefsHeader->crc, data, size);
	return FSWrite (state->fileRefNum, &size, data);
}

pascal OSErr WritePrefsCollectionToFile (PrefsCollection pc, const FSSpec *fssP)
{
	OSErr err = noErr;

	short fileRefNum;

	if (!(err = FSpOpenDF (fssP,fsWrPerm,&fileRefNum)))
	{
		if (!(err = SetEOF (fileRefNum, sizeof (tPrefsFormat))))
		if (!(err = SetFPos (fileRefNum, fsFromStart, sizeof (tPrefsFormat))))
		{
			CollectionFlattenUPP upp = NewCollectionFlattenProc (CollectionFlattenProc);

			if (!upp)
				err = MemError ( );
			else
			{
				tPrefsFormat				header	= { 0, kPrefsFileMagicCookie, kPrefsFileMagicCookie };
				tCollectionFlattenState		state	= { fileRefNum, &header };

				if (!(err = FlattenCollection (Collection (pc), upp, &state)))
				{
					if (!(err = SetFPos (fileRefNum, fsFromStart, 0)))
					{
						Size writeSize = sizeof (header);
						err = FSWrite (fileRefNum,&writeSize,&header);
					}
				}

				DisposeRoutineDescriptor (upp);
			}
		}

		OSErr err2 = FSClose (fileRefNum);
		if (!err) err = err2;
		err2 = FlushVol (nil, fssP->vRefNum);
		if (!err) err = err2;
	}

	return err;
}

pascal OSErr SetPreference (PrefsCollection pc, PrefsType pt, PrefsID pid, Size ps, void *pb)
{
	if (!MoreAssert (kPrefsReservedType != pt)) return paramErr;

	OSErr err = RemoveCollectionItem (Collection (pc), pt, pid);

	if (!err || err == collectionItemNotFoundErr)
		err = AddCollectionItem (Collection (pc), pt, pid, ps, pb);

	return err;
}

pascal OSErr GetPreference (PrefsCollection pc, PrefsType pt, PrefsID pid, Size ps, Size *aps, void *pb)
{
	if (!MoreAssert (kPrefsReservedType != pt))
		return paramErr;
	else
	{
		if (aps) *aps = ps; else aps = &ps;
		return GetCollectionItem (Collection (pc), pt, pid, aps, pb);
	}
}

pascal OSErr PrefsAreSame (PrefsCollection,PrefsCollection, Boolean *areSame)
{
	DebugStr ("\pno comparison performed -- returning false");
	*areSame = false;
	return noErr;
}

pascal OSErr GetPreferenceSize (PrefsCollection pc, PrefsType pt, PrefsID pid, Size *ps)
{
	if (!MoreAssert (kPrefsReservedType != pt))
		return paramErr;
	else if (!MoreAssert (pc && ps))
		return paramErr;
	else
		return GetCollectionItem (Collection (pc), pt, pid, ps, nil);
}
